@@ -17,7 +17,7 @@
 %                                 April 2007                                  %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2021 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright @ 1999 ImageMagick Studio LLC, a non-profit organization         %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
@@ -60,14 +60,31 @@
 #include "MagickCore/resource_.h"
 #include "MagickCore/utility.h"
 #if defined(MAGICKCORE_OPENEXR_DELEGATE)
-#include <ImfCRgbaFile.h>
-#if IMF_VERSION_NUMBER > 1
 #include <OpenEXRConfig.h>
+#include <ImfCRgbaFile.h>
+/*
+  Version 3.1.0 was released on 2021-07-21
+  We can remove backwards compatibility in a couple years.
+*/
+#define OPENEXR_VERSION_HACK  \
+    (((OPENEXR_VERSION_MAJOR) << 24) |  \
+     ((OPENEXR_VERSION_MINOR) << 16) |  \
+     ((OPENEXR_VERSION_PATCH) << 8))
+#if (OPENEXR_VERSION_HACK) >= 0x3010000
+#define USE_OPENEXR_CORE
+#include <openexr.h>
 #endif
 
 /*
-  Typedef declaractions.
+  Typedef declarations.
 */
+#if defined(USE_OPENEXR_CORE)
+typedef struct _EXRUserData
+{
+  ExceptionInfo
+    *exception;
+} EXRUserData;
+#else
 typedef struct _EXRWindowInfo
 {
   int
@@ -76,6 +93,7 @@ typedef struct _EXRWindowInfo
     min_x,
     min_y;
 } EXRWindowInfo;
+#endif
 
 /*
   Forward declarations.
@@ -145,6 +163,309 @@ static MagickBooleanType IsEXR(const unsigned char *magick,const size_t length)
 %    o exception: return any errors or warnings in this structure.
 %
 */
+
+#if defined(USE_OPENEXR_CORE)
+static void EXRError(exr_const_context_t ctxt,exr_result_t magick_unused(code),
+  const char* msg)
+{
+  EXRUserData
+    *user_data;
+
+  magick_unreferenced(code);
+  exr_get_user_data(ctxt,(void **) &user_data);
+  (void) ThrowMagickException(user_data->exception,GetMagickModule(),
+    CoderError,"CorruptImage","`%s'",msg);
+}
+
+static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
+{
+  exr_attr_box2i_t
+    data_window;
+  
+  exr_chunk_info_t
+    cinfo;
+
+  exr_coding_channel_info_t
+    *channel;
+
+  exr_compression_t
+    compression = EXR_COMPRESSION_NONE;
+
+  exr_context_t
+    ctxt;
+
+  exr_context_initializer_t
+    initializer = EXR_DEFAULT_CONTEXT_INITIALIZER;
+
+  exr_decode_pipeline_t
+    decoder = EXR_DECODE_PIPELINE_INITIALIZER;
+
+  exr_result_t
+    result;
+
+  exr_storage_t
+    storage_type = EXR_STORAGE_LAST_TYPE;
+
+  EXRUserData
+    user_data = { 0 };
+
+  Image
+    *image;
+
+  int
+    part_index = 0;
+  
+  int16_t
+    c;
+
+  int32_t
+    scans_per_chunk;
+
+  MagickBooleanType
+    status;
+
+  PixelChannel
+    pixel_channels[MaxPixelChannels] = { UndefinedPixelChannel };
+
+  size_t
+    pixel_count,
+    pixel_size,
+    number_meta_channels;
+
+  ssize_t
+    y;
+  
+  uint8_t
+    *chunk,
+    *p;
+
+  /*
+    Open image.
+  */
+  assert(image_info != (const ImageInfo *) NULL);
+  assert(image_info->signature == MagickCoreSignature);
+  assert(exception != (ExceptionInfo *) NULL);
+  assert(exception->signature == MagickCoreSignature);
+  if (IsEventLogging() != MagickFalse)
+    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",
+      image_info->filename);
+  image=AcquireImage(image_info,exception);
+  status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
+  if (status == MagickFalse)
+    {
+      image=DestroyImageList(image);
+      return((Image *) NULL);
+    }
+  user_data.exception=exception;
+  initializer.error_handler_fn=EXRError;
+  initializer.user_data=(void *) &user_data;
+  initializer.max_image_height=(int) MagickMin(INT_MAX,
+    GetMagickResourceLimit(HeightResource));
+  initializer.max_image_width=(int) MagickMin(INT_MAX,
+    GetMagickResourceLimit(WidthResource));
+  result=exr_start_read(&ctxt,image->filename,&initializer);
+  if (result != EXR_ERR_SUCCESS)
+    {
+      exr_finish(&ctxt);
+      ThrowReaderException(BlobError,"UnableToOpenBlob");
+    }
+  result=exr_get_data_window(ctxt,part_index,&data_window);
+  if (result == EXR_ERR_SUCCESS)
+    result=exr_get_compression(ctxt,part_index,&compression);
+  if (result == EXR_ERR_SUCCESS)
+    result=exr_get_storage(ctxt,part_index,&storage_type);
+  if ((result != EXR_ERR_SUCCESS) ||
+      (data_window.min.x >= data_window.max.x) ||
+      (data_window.min.y >= data_window.max.y))
+    {
+      exr_finish(&ctxt);
+      ThrowReaderException(CorruptImageError,"CorruptImage");
+    }
+  if (storage_type != EXR_STORAGE_SCANLINE)
+    {
+      exr_finish(&ctxt);
+      (void) ThrowMagickException(exception,GetMagickModule(),
+        CorruptImageError,"Unsupported storage type","`%d'",
+        (int) storage_type);
+      image=DestroyImageList(image);
+      (void) CloseBlob(image);
+      return((Image *) NULL);
+    }
+  image->columns=(size_t) (data_window.max.x-data_window.min.x+1L);
+  image->rows=(size_t) (data_window.max.y-data_window.min.y+1L);
+  image->gamma=1.0;
+  image->compression=NoCompression;
+  switch (compression)
+  {
+    case EXR_COMPRESSION_RLE: image->compression=RLECompression; break;
+    case EXR_COMPRESSION_ZIPS: image->compression=ZipSCompression; break;
+    case EXR_COMPRESSION_ZIP: image->compression=ZipCompression; break;
+    case EXR_COMPRESSION_PIZ: image->compression=PizCompression; break;
+    case EXR_COMPRESSION_PXR24: image->compression=Pxr24Compression; break;
+    case EXR_COMPRESSION_B44: image->compression=B44Compression; break;
+    case EXR_COMPRESSION_B44A: image->compression=B44ACompression; break;
+    case EXR_COMPRESSION_DWAA: image->compression=DWAACompression; break;
+    case EXR_COMPRESSION_DWAB: image->compression=DWABCompression; break;
+    default: break;
+  }
+  if (image_info->ping != MagickFalse)
+    {
+      exr_finish(&ctxt);
+      (void) CloseBlob(image);
+      return(GetFirstImageInList(image));
+    }
+  status=SetImageExtent(image,image->columns,image->rows,exception);
+  if (status == MagickFalse)
+    {
+      exr_finish(&ctxt);
+      (void) CloseBlob(image);
+      return(DestroyImageList(image));
+    }
+  result=exr_get_scanlines_per_chunk(ctxt,part_index,&scans_per_chunk);
+  if (result == EXR_ERR_SUCCESS)
+    result=exr_read_scanline_chunk_info(ctxt,part_index,data_window.min.y,
+      &cinfo);
+  if (result == EXR_ERR_SUCCESS)
+    result=exr_decoding_initialize(ctxt,part_index,&cinfo,&decoder);
+  if (result != EXR_ERR_SUCCESS)
+    {
+      exr_finish(&ctxt);
+      ThrowReaderException(CorruptImageError,"CorruptImage");
+    }
+  if (decoder.channel_count >= (MaxPixelChannels-MetaPixelChannels))
+    {
+      exr_decoding_destroy(ctxt,&decoder);
+      exr_finish(&ctxt);
+      ThrowReaderException(CorruptImageError,"MaximumChannelsExceeded");
+    }
+  channel=decoder.channels;
+  pixel_size=0;
+  number_meta_channels=0;
+  for (c = 0; c < decoder.channel_count; ++c)
+  {
+    char
+      channel_name[MagickPathExtent];
+
+    if ((channel->data_type != EXR_PIXEL_HALF) &&
+        (channel->data_type != EXR_PIXEL_FLOAT))
+      {
+        status=MagickFalse;
+        (void) ThrowMagickException(exception,GetMagickModule(),
+          CorruptImageError,"Unsupported channel data type","`%d'",
+          (int) channel->data_type );
+        break;
+      }
+    pixel_size+=(size_t) channel->bytes_per_element;
+    if (LocaleNCompare(channel->channel_name,"R",1) == 0)
+      pixel_channels[c]=RedPixelChannel;
+    else if (LocaleNCompare(channel->channel_name,"G",1) == 0)
+      pixel_channels[c]=GreenPixelChannel;
+    else if (LocaleNCompare(channel->channel_name,"B",1) == 0)
+      pixel_channels[c]=BluePixelChannel;
+    else if (LocaleNCompare(channel->channel_name,"A",1) == 0)
+      {
+        pixel_channels[c]=AlphaPixelChannel;
+        image->alpha_trait=BlendPixelTrait;
+      }
+    else if (LocaleNCompare(channel->channel_name,"Y",1) == 0)
+      pixel_channels[c]=IndexPixelChannel; /* Gray channel */
+    else
+      {
+        pixel_channels[c]=(PixelChannel) (MetaPixelChannels+
+          number_meta_channels);
+        FormatLocaleString(channel_name,MagickPathExtent,
+          "exr:meta-channel.%d",(int) number_meta_channels);
+        number_meta_channels++;
+        SetImageProperty(image,channel_name,channel->channel_name,exception);
+      }
+    channel++;
+  }
+  if (status != MagickFalse)
+    status=SetPixelMetaChannels(image,number_meta_channels,exception);
+  if (status == MagickFalse)
+    {
+      exr_decoding_destroy(ctxt,&decoder);
+      exr_finish(&ctxt);
+      (void) CloseBlob(image);
+      return(DestroyImageList(image));
+    }
+  pixel_count=(size_t) scans_per_chunk*image->columns;
+  chunk=(uint8_t *) AcquireQuantumMemory(pixel_size,pixel_count);
+  memset(chunk,0,pixel_size*pixel_count);
+  p=chunk;
+  channel=decoder.channels;
+  for (c = 0; c < decoder.channel_count; ++c)
+  {
+    channel->decode_to_ptr=p;
+    channel->user_pixel_stride=(int32_t) pixel_size;
+    channel->user_line_stride=(int32_t) (image->columns*pixel_size);
+    p+=channel->bytes_per_element;
+    channel++;
+  }
+  for (y=0; y < (ssize_t) image->rows; y+= (ssize_t) scans_per_chunk)
+  {
+    Quantum
+      *q;
+
+    ssize_t
+      x;
+
+    if (y == 0)
+      result=exr_decoding_choose_default_routines(ctxt,part_index,&decoder);
+    else
+      {
+        int
+          yy = (int) y + data_window.min.y;
+
+        result=exr_read_scanline_chunk_info(ctxt,part_index,yy,&cinfo);
+        if (result == EXR_ERR_SUCCESS)
+          result=exr_decoding_update(ctxt,part_index,&cinfo,&decoder);
+      }
+    if (result == EXR_ERR_SUCCESS)
+      result=exr_decoding_run(ctxt,part_index,&decoder);
+    if (result != EXR_ERR_SUCCESS)
+      break;
+    q=QueueAuthenticPixels(image,0,y,image->columns,(size_t) scans_per_chunk,
+      exception);
+    if (q == (Quantum *) NULL)
+      break;
+    p=chunk;
+    for (x=0; x < (ssize_t) pixel_count; x++)
+    {
+      channel=decoder.channels;
+      for (c = 0; c < decoder.channel_count; ++c)
+      {
+        float
+          value;
+
+        PixelChannel
+          pixel_channel;
+
+        if (channel->data_type == EXR_PIXEL_HALF)
+          value=HalfToSinglePrecision(*(unsigned short *) p);
+        else
+          value=*(float *) p;
+        pixel_channel=pixel_channels[c];
+        if (pixel_channel == IndexPixelChannel)
+          SetPixelGray(image,ClampToQuantum((double) QuantumRange*value),q);
+        else
+          SetPixelChannel(image,pixel_channel,ClampToQuantum(
+            (double) QuantumRange*value),q);
+        p+=channel->bytes_per_element;
+        channel++;
+      }
+      q+=GetPixelChannels(image);
+    }
+  }
+  chunk=(uint8_t *) RelinquishMagickMemory(chunk);
+  exr_decoding_destroy(ctxt,&decoder);
+  exr_finish(&ctxt);
+  (void) CloseBlob(image);
+  if (result != EXR_ERR_SUCCESS)
+    return(DestroyImageList(image));
+  return(GetFirstImageInList(image));
+}
+#else
 static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
 {
   const ImfHeader
@@ -183,11 +504,11 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
   */
   assert(image_info != (const ImageInfo *) NULL);
   assert(image_info->signature == MagickCoreSignature);
-  if (image_info->debug != MagickFalse)
-    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",
-      image_info->filename);
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickCoreSignature);
+  if (IsEventLogging() != MagickFalse)
+    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",
+      image_info->filename);
   image=AcquireImage(image_info,exception);
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
   if (status == MagickFalse)
@@ -206,8 +527,8 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
   hdr_info=ImfInputHeader(file);
   ImfHeaderDataWindow(hdr_info,&display_window.min_x,&display_window.min_y,
     &display_window.max_x,&display_window.max_y);
-  image->columns=((size_t) display_window.max_x-display_window.min_x+1UL);
-  image->rows=((size_t) display_window.max_y-display_window.min_y+1UL);
+  image->columns=(size_t) (display_window.max_x-display_window.min_x+1L);
+  image->rows=(size_t) (display_window.max_y-display_window.min_y+1L);
   image->alpha_trait=BlendPixelTrait;
   (void) SetImageColorspace(image,RGBColorspace,exception);
   image->gamma=1.0;
@@ -253,7 +574,7 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
     }
   ImfHeaderDataWindow(hdr_info,&data_window.min_x,&data_window.min_y,
     &data_window.max_x,&data_window.max_y);
-  columns=((size_t) data_window.max_x-data_window.min_x+1UL);
+  columns=(size_t) (data_window.max_x-data_window.min_x+1L);
   if ((display_window.min_x > data_window.max_x) ||
       (display_window.min_x+(int) image->columns <= data_window.min_x))
     scanline=(ImfRgba *) NULL;
@@ -292,8 +613,7 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
         continue;
       }
     (void) memset(scanline,0,columns*sizeof(*scanline));
-    if (ImfInputSetFrameBuffer(file,scanline-data_window.min_x-columns*yy,1,
-      columns) == 0)
+    if (ImfInputSetFrameBuffer(file,scanline-data_window.min_x-(ssize_t) columns*yy,1,columns) == 0)
       {
         status=MagickFalse;
         break;
@@ -313,14 +633,14 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
         SetPixelViaPixelInfo(image,&image->background_color,q);
       else
         {
-          SetPixelRed(image,ClampToQuantum((MagickRealType) QuantumRange*
-            ImfHalfToFloat(scanline[xx].r)),q);
-          SetPixelGreen(image,ClampToQuantum((MagickRealType) QuantumRange*
-            ImfHalfToFloat(scanline[xx].g)),q);
-          SetPixelBlue(image,ClampToQuantum((MagickRealType) QuantumRange*
-            ImfHalfToFloat(scanline[xx].b)),q);
-          SetPixelAlpha(image,ClampToQuantum((MagickRealType) QuantumRange*
-            ImfHalfToFloat(scanline[xx].a)),q);
+          SetPixelRed(image,ClampToQuantum((double) QuantumRange*
+            (double) ImfHalfToFloat(scanline[xx].r)),q);
+          SetPixelGreen(image,ClampToQuantum((double) QuantumRange*
+            (double) ImfHalfToFloat(scanline[xx].g)),q);
+          SetPixelBlue(image,ClampToQuantum((double) QuantumRange*
+            (double) ImfHalfToFloat(scanline[xx].b)),q);
+          SetPixelAlpha(image,ClampToQuantum((double) QuantumRange*
+            (double) ImfHalfToFloat(scanline[xx].a)),q);
         }
       q+=GetPixelChannels(image);
     }
@@ -331,10 +651,14 @@ static Image *ReadEXRImage(const ImageInfo *image_info,ExceptionInfo *exception)
   (void) ImfCloseInputFile(file);
   if (status == MagickFalse)
     ThrowReaderException(CorruptImageError,"UnableToReadImageData");
-  (void) CloseBlob(image);
+  if (CloseBlob(image) == MagickFalse)
+    status=MagickFalse;
+  if (status == MagickFalse)
+    return(DestroyImageList(image));
   return(GetFirstImageInList(image));
 }
 #endif
+#endif
 
 /*
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -485,10 +809,10 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image,
   assert(image_info->signature == MagickCoreSignature);
   assert(image != (Image *) NULL);
   assert(image->signature == MagickCoreSignature);
-  if (image->debug != MagickFalse)
-    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickCoreSignature);
+  if (IsEventLogging() != MagickFalse)
+    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   status=OpenBlob(image_info,image,WriteBinaryBlobMode,exception);
   if (status == MagickFalse)
     return(status);
@@ -634,20 +958,23 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image,
       break;
     for (x=0; x < (ssize_t) image->columns; x++)
     {
-      ImfFloatToHalf(QuantumScale*GetPixelRed(image,p),&half_quantum);
+      ImfFloatToHalf(QuantumScale*(double) GetPixelRed(image,p),&half_quantum);
       scanline[x].r=half_quantum;
-      ImfFloatToHalf(QuantumScale*GetPixelGreen(image,p),&half_quantum);
+      ImfFloatToHalf(QuantumScale*(double) GetPixelGreen(image,p),
+        &half_quantum);
       scanline[x].g=half_quantum;
-      ImfFloatToHalf(QuantumScale*GetPixelBlue(image,p),&half_quantum);
+      ImfFloatToHalf(QuantumScale*(double) GetPixelBlue(image,p),&half_quantum);
       scanline[x].b=half_quantum;
-      if (image->alpha_trait == UndefinedPixelTrait)
+      if ((image->alpha_trait & BlendPixelTrait) == 0)
         ImfFloatToHalf(1.0,&half_quantum);
       else
-        ImfFloatToHalf(QuantumScale*GetPixelAlpha(image,p),&half_quantum);
+        ImfFloatToHalf(QuantumScale*(double) GetPixelAlpha(image,p),
+          &half_quantum);
       scanline[x].a=half_quantum;
       p+=GetPixelChannels(image);
     }
-    ImfOutputSetFrameBuffer(file,scanline-(y*image->columns),1,image->columns);
+    ImfOutputSetFrameBuffer(file,scanline-(y*(ssize_t) image->columns),1,
+      image->columns);
     ImfOutputWritePixels(file,1);
   }
   (void) ImfCloseOutputFile(file);
@@ -655,7 +982,8 @@ static MagickBooleanType WriteEXRImage(const ImageInfo *image_info,Image *image,
   (void) FileToImage(image,write_info->filename,exception);
   (void) RelinquishUniqueFileResource(write_info->filename);
   write_info=DestroyImageInfo(write_info);
-  (void) CloseBlob(image);
-  return(MagickTrue);
+  if (CloseBlob(image) == MagickFalse)
+    status=MagickFalse;
+  return(status);
 }
 #endif
